# Copyright (C) 2017-2018  Lars Wirzenius
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import ick2


class ProjectAPI(ick2.ResourceApiBase):

    def __init__(self, state):
        super().__init__('projects', state)
        self._pi = ick2.PipelineInstances(self.get_state())

    def mangle_new_resource(self, resource):
        new = dict(resource)
        if 'next_build_id' not in new:
            new['next_build_id'] = None
        return new

    def mangle_updated_resource(self, old, new):
        new = dict(new)
        new['next_build_id'] = old.get('next_build_id')
        return new

    def get_resource_name(self, resource):
        return resource['project']

    def get_routes(self, path):  # pragma: no cover
        return super().get_routes(path) + self.get_pipeline_routes(path)

    def get_pipeline_routes(self, path):  # pragma: no cover
        pipeline_path = '{}/<project>/pipelines/<pipeline>'.format(path)
        return [
            {
                'method': 'GET',
                'path': pipeline_path,
                'callback': self.GET(self.get_pipeline),
            },
            {
                'method': 'PUT',
                'path': pipeline_path,
                'callback': self.PUT(self.set_pipeline_callback),
            },
            {
                'needs-authorization': False,
                'method': 'GET',
                'path': pipeline_path + '/+trigger',
                'callback': self.GET(self.trigger_pipeline),
            },
        ]

    def get_pipeline(self, project, pipeline, **kwargs):
        p = self._state.get_resource(self._type_name, project)
        if pipeline not in p['pipelines']:
            raise ick2.NotFound()

        pl = self._pi.get_instance(project, pipeline)
        return {
            'status': pl.get('status', 'idle'),
        }

    def set_pipeline_callback(
            self, body, project, pipeline, **kwargs):  # pragma: no cover
        return self.set_pipeline(body['status'], project, pipeline)

    def set_pipeline(self, status, project, pipeline):
        ick2.log.log(
            'trace', msg_text='Setting pipeline status',
            project=project, pipeline=pipeline, status=status)

        allowed_changes = {
            'idle': 'triggered',
            'triggered': 'building',
            'building': 'idle',
        }

        p = self._state.get_resource(self._type_name, project)
        if pipeline not in p['pipelines']:
            ick2.log.log(
                'error', msg_text='Project not found', project=project)
            raise ick2.NotFound()
        ick2.log.log('trace', msg_text='Found project', project=p)

        pl = self._pi.get_instance(project, pipeline)
        old_status = pl.get('status', 'idle')
        if allowed_changes[old_status] != status:
            raise ick2.WrongPipelineStatus(status)
        pl['status'] = status
        self._pi.update_instance(project, pipeline, pl)
        return {'status': status}

    # This needs to go away as it is not protected. Once an IDP is
    # added.
    def trigger_pipeline(
            self, project, pipeline, **kwargs):  # pragma: no cover
        return self.set_pipeline('triggered', project, pipeline)
